import asyncio
import logging
import time
from datetime import datetime, timedelta, timezone
from operator import attrgetter
from pathlib import Path
from typing import Any, Final, Optional, Union

import aiojobs
import attr
import yaml
from aiojobs_protocols import SchedulerProtocol
from aiotgbot import (Bot, BotBlocked, BotUpdate, Chat, InlineKeyboardButton,
                      InlineKeyboardMarkup, Message, ParseMode)
from aiotgbot.api_types import (InputMediaAudio, InputMediaDocument,
                                InputMediaPhoto, InputMediaVideo, User)
from more_itertools import chunked

ALBUM_WAIT_TIMEOUT = 1  # seconds
CHAT_LIST_KEY: Final[str] = 'chat_list'
REPLY_PREFIX: Final[str] = 'reply'
CONFIG_KEY: Final['str'] = 'config'

logger = logging.getLogger('feedback-bot')


def get_software() -> str:
    from aiotgbot.helpers import get_python_version  # isort:skip
    from aiotgbot import __version__ as aiotgbot_version  # isort:skip
    from . import __version__  # isort:skip
    return (f'Python/{get_python_version()} '
            f'aiotgbot/{aiotgbot_version} '
            f'feedback-bot/{__version__}')


def path(_str: str) -> Path:
    return Path(_str)


def debug() -> bool:
    return __debug__


def user_name(user_chat: Union[User, Chat]) -> str:
    if user_chat.first_name is None:
        raise RuntimeError('First name of private chat must be not empty')
    if user_chat.last_name is not None:
        return f'{user_chat.first_name} {user_chat.last_name}'
    else:
        return user_chat.first_name


def user_link(user_chat: Union[User, Chat]) -> str:
    return f'<a href="tg://user?id={user_chat.id}">{user_name(user_chat)}</a>'


def chat_key(chat_id: int) -> str:
    return f'chat|{chat_id}'


async def set_chat(bot: Bot, key: str, chat: Optional[Chat] = None) -> None:
    await bot.storage.set(key, chat.to_dict() if chat is not None else None)


async def get_chat(bot: Bot, key: str) -> Optional[Chat]:
    data = await bot.storage.get(key)
    if isinstance(data, dict):
        return Chat.from_dict(data)
    elif data is None:
        return None
    else:
        raise RuntimeError('Chat data is not dict or None')


async def get_chat_list(bot: Bot) -> list[Chat]:
    chat_list = await bot.storage.get(CHAT_LIST_KEY)
    if chat_list is None:
        raise RuntimeError('Chat list not in storage')
    assert isinstance(chat_list, list)
    assert all(isinstance(item, dict) for item in chat_list)
    return [Chat.from_dict(item) for item in chat_list]


async def set_chat_list(bot: Bot, chat_list: list[Chat]) -> None:
    await bot.storage.set(
        CHAT_LIST_KEY,
        [chat.to_dict() for chat in chat_list]
    )


async def add_chat_to_list(bot: Bot, chat: Chat) -> None:
    config = bot[CONFIG_KEY]
    assert isinstance(config, Config)
    chat_list = await get_chat_list(bot)
    if all(item.id != chat.id for item in chat_list):
        chat_list.append(chat)
        if len(chat_list) > config.chat_list_size:
            chat_list.pop(0)
        await set_chat_list(bot, chat_list)


async def remove_chat_from_list(bot: Bot, remove_id: int) -> None:
    chat_list = await get_chat_list(bot)
    chat_list = [chat for chat in chat_list if chat.id != remove_id]
    await set_chat_list(bot, chat_list)


async def send_from_message(bot: Bot, chat_id: int, from_chat: Chat) -> None:
    await bot.send_message(chat_id, f'От {user_link(from_chat)}',
                           parse_mode=ParseMode.HTML)


async def reply_menu(bot: Bot, chat_id: Union[int, str]) -> None:
    chat_list = await get_chat_list(bot)
    if len(chat_list) == 0:
        await bot.send_message(chat_id, 'Некому отвечать.')
    else:
        await bot.send_message(
            chat_id, 'Выберите пользователя для ответа.',
            reply_markup=InlineKeyboardMarkup(
                inline_keyboard=[[InlineKeyboardButton(
                    user_name(chat), callback_data=f'{REPLY_PREFIX}|{chat.id}'
                ) for chat in chunk] for chunk in chunked(chat_list, 2)]
            )
        )


async def send_user_message(bot: Bot, message: Message) -> None:
    assert 'album_forwarder' in bot
    assert isinstance(bot['album_forwarder'], AlbumForwarder)
    current_chat = await get_chat(bot, 'current_chat')
    if current_chat is None and message.media_group_id is not None:
        await bot['album_forwarder'].add_message(message)
        logger.debug('Add next media group item to forwarder')
        return
    if current_chat is None:
        await bot.send_message(message.chat.id, 'Нет текущего пользователя')
        logger.debug('Skip message to user: no current user')
        return

    stopped = await Stopped.get(bot, current_chat.id)
    if stopped is not None:
        await bot.send_message(
            message.chat.id,
            f'{user_link(current_chat)} меня заблокировал '
            f'{stopped.dt:%Y-%m-%d %H:%M:%S %Z}.',
            parse_mode=ParseMode.HTML)
        return

    if message.media_group_id is not None:
        await bot['album_forwarder'].add_message(message, current_chat.id)
        logger.debug('Add first media group item to forwarder')
        return

    logger.debug('Send message to "%s"', current_chat.to_dict())
    try:
        await bot.copy_message(current_chat.id, message.chat.id,
                               message.message_id)
    except BotBlocked:
        await remove_chat_from_list(bot, current_chat.id)
        await Stopped(blocked=True).set(bot, current_chat.id)
        await bot.send_message(
            message.chat.id, f'{user_link(current_chat)} меня заблокировал.',
            parse_mode=ParseMode.HTML, disable_web_page_preview=True)
        logger.info('Blocked by user "%s"', current_chat.to_dict())
        return
    else:
        await bot.send_message(
            message.chat.id,
            f'Сообщение отправлено {user_link(current_chat)}.',
            parse_mode=ParseMode.HTML, disable_web_page_preview=True)


@attr.s(slots=True, frozen=True, auto_attribs=True)
class FromUserFilter:

    async def check(self, bot: Bot, update: BotUpdate) -> bool:  # noqa
        config = bot[CONFIG_KEY]
        assert isinstance(config, Config)

        return (update.message is not None and
                update.message.from_ is not None and
                update.message.from_.username != config.admin_username)


@attr.s(slots=True, frozen=True, auto_attribs=True)
class FromAdminFilter:

    async def check(self, bot: Bot, update: BotUpdate) -> bool:  # noqa
        config = bot[CONFIG_KEY]
        assert isinstance(config, Config)

        return (update.message is not None and
                update.message.from_ is not None and
                update.message.from_.username == config.admin_username)


class AlbumForwarder:
    __slots__ = '_queues', '_scheduler', '_bot'

    def __init__(self, bot: Bot) -> None:
        self._queues: 'Final[dict[str, asyncio.Queue[Message]]]' = {}
        self._scheduler: Optional[SchedulerProtocol] = None
        self._bot: Final[Bot] = bot

    async def add_message(self, message: Message,
                          chat_id: Optional[int] = None,
                          add_from_info: bool = False) -> None:
        if self._scheduler is None:
            raise RuntimeError('Album forwarder not started')
        if message.media_group_id is None:
            raise RuntimeError('Message in album must have media_group_id')
        if message.media_group_id in self._queues:
            self._queues[message.media_group_id].put_nowait(message)
        elif chat_id is not None:
            self._queues[message.media_group_id] = asyncio.Queue()
            self._queues[message.media_group_id].put_nowait(message)
            await self._scheduler.spawn(
                self._send(message.media_group_id, chat_id, add_from_info))
        else:
            logger.warning('Skip media group item as latecomer %s', message)

    async def _send(self, media_group_id: str, chat_id: int,
                    add_from_info: bool = False) -> None:
        assert media_group_id in self._queues
        media: list[Union[InputMediaAudio, InputMediaPhoto, InputMediaVideo,
                          InputMediaDocument]] = []
        from_chat: Optional[Chat] = None
        message_count: int = 0
        while True:
            try:
                message = await asyncio.wait_for(
                    self._queues[media_group_id].get(),
                    timeout=ALBUM_WAIT_TIMEOUT)
            except asyncio.TimeoutError:
                break
            assert isinstance(message, Message)
            message_count += 1
            from_chat = message.chat
            if message.audio is not None:
                media.append(InputMediaAudio(
                    media=message.audio.file_id,
                    caption=message.caption,
                    caption_entities=message.caption_entities,
                    duration=message.audio.duration,
                    performer=message.audio.performer,
                    title=message.audio.title
                ))
            elif message.document is not None:
                media.append(InputMediaDocument(
                    media=message.document.file_id,
                    caption=message.caption,
                    caption_entities=message.caption_entities
                ))
            elif message.photo is not None:
                media.append(InputMediaPhoto(
                    media=max(message.photo,
                              key=attrgetter('file_size')).file_id,
                    caption=message.caption,
                    caption_entities=message.caption_entities
                ))
            elif message.video is not None:
                media.append(InputMediaVideo(
                    media=message.video.file_id,
                    caption=message.caption,
                    caption_entities=message.caption_entities,
                    width=message.video.width,
                    height=message.video.height,
                    duration=message.video.duration
                ))
        if len(media) > 0:
            assert from_chat is not None
            if add_from_info:
                await send_from_message(self._bot, chat_id, from_chat)
            await self._bot.send_media_group(chat_id, media)
            await self._bot.send_message(
                from_chat.id, f'Переслано элементов группы: {len(media)}')
            logger.debug('Forwarded %d media group items', len(media))
        elif from_chat is not None:
            await self._bot.send_message(
                from_chat.id, 'Не удалось переслать элементов '
                              f'неподдерживаемого типа: {message_count}')
            logger.debug('Failed to forward %d media group items of '
                         'unsupported type', message_count)
        else:
            logger.debug('No media group items to forward')
        self._queues.pop(media_group_id)

    async def start(self) -> None:
        self._scheduler = await aiojobs.create_scheduler(
            close_timeout=float('inf'),
            exception_handler=self._scheduler_exception_handler)

    async def stop(self) -> None:
        if self._scheduler is None:
            raise RuntimeError('Album forwarder not started')
        await self._scheduler.close()
        assert len(self._queues) == 0

    @staticmethod
    def _scheduler_exception_handler(_: SchedulerProtocol,
                                     context: dict[str, Any]) -> None:
        logger.exception('Album forward error', exc_info=context['exception'])


def _now_with_tz() -> datetime:
    return datetime.now(timezone(timedelta(seconds=-time.timezone)))


@attr.s(slots=True, frozen=True, auto_attribs=True)
class Stopped:
    dt: datetime = attr.ib(factory=_now_with_tz)
    blocked: bool = False

    @staticmethod
    def _key(chat_id: int) -> str:
        return f'stopped|{chat_id}'

    async def set(self, bot: Bot, chat_id: int) -> None:
        await bot.storage.set(
            self._key(chat_id),
            {'dt': self.dt.isoformat(), 'error': self.blocked}
        )

    @staticmethod
    async def get(bot: Bot, chat_id: int) -> Optional['Stopped']:
        data = await bot.storage.get(Stopped._key(chat_id))
        if data is not None:
            assert isinstance(data, dict)
            return Stopped(datetime.fromisoformat(data['dt']), data['error'])
        else:
            return None

    @staticmethod
    async def delete(bot: Bot, chat_id: int) -> None:
        await bot.storage.delete(Stopped._key(chat_id))


@attr.s(auto_attribs=True, frozen=True)
class Config:
    admin_username: str
    tg_token: str
    chat_list_size: int

    @staticmethod
    def load(config_path: Path) -> 'Config':
        with config_path.open('r') as file:
            data = yaml.load(file, yaml.CSafeLoader)
            return Config(**data)



import asyncio
import logging
import re
from pathlib import Path
from typing import AsyncIterator, Final

from aiotgbot import (Bot, BotUpdate, ContentType, GroupChatFilter,
                      HandlerTable, ParseMode, PollBot, PrivateChatFilter,
                      Runner, TelegramError)
from aiotgbot.api_types import (BotCommand, BotCommandScopeAllPrivateChats,
                                BotCommandScopeChat)
from aiotgbot.storage_sqlite import SQLiteStorage

from .helpers import (CHAT_LIST_KEY, CONFIG_KEY, REPLY_PREFIX, AlbumForwarder,
                      Config, FromAdminFilter, FromUserFilter, Stopped,
                      add_chat_to_list, chat_key, debug, get_chat,
                      get_software, path, remove_chat_from_list, reply_menu,
                      send_from_message, send_user_message, set_chat,
                      user_link)

SOFTWARE: Final[str] = get_software()
USER_COMMANDS: Final[tuple[BotCommand, ...]] = (
    BotCommand('help', 'Помощь'),
    BotCommand('stop', 'Остановить и не получать больше сообщения')
)
ADMIN_COMMANDS: Final[tuple[BotCommand, ...]] = (
    BotCommand('help', 'Помощь'),
    BotCommand('reply', 'Ответить пользователю'),
    BotCommand('add_to_group', 'Добавить в группу'),
    BotCommand('remove_from_group', 'Удалить из группы'),
    BotCommand('reset', 'Сбросить состояние')
)
GROUP_COMMANDS: Final[tuple[BotCommand, ...]] = (
    BotCommand('help', 'Помощь'),
    BotCommand('reply', 'Ответить пользователю')
)
CHAT_ID_GROUP: Final[str] = 'chat_id'
REPLY_RXP: 'Final[re.Pattern[str]]' = re.compile(
    rf'^{REPLY_PREFIX}\|(?P<{CHAT_ID_GROUP}>-?\d+)$')
ALBUM_FORWARDER_KEY: Final[str] = 'album_forwarder'
GROUP_CHAT_KEY: Final[str] = 'group_chat'
ADMIN_CHAT_ID_KEY: Final[str] = 'admin_chat_id'
CURRENT_CHAT_KEY: Final[str] = 'current_chat'
WAIT_REPLY_FROM_ID_KEY: Final[str] = 'wait_reply_from_id'
TZ_KEY: Final['str'] = 'TZ'


logger = logging.getLogger('feedback-bot')
handlers = HandlerTable()


@handlers.message(commands=['start'],
                  filters=[PrivateChatFilter(), FromUserFilter()])
async def user_start_command(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    assert update.message.from_ is not None
    logger.info('Start command from "%s"', update.message.from_.to_dict())
    stopped = await Stopped.get(bot, update.message.chat.id)
    if stopped is not None:
        await Stopped.delete(bot, update.message.chat.id)
        await bot.send_message(update.message.chat.id, 'С возвращением!')
    await set_chat(bot, chat_key(update.message.chat.id), update.message.chat)
    await bot.send_message(update.message.chat.id,
                           'Пришлите сообщение или задайте вопрос. '
                           'Также вы можете использовать следующие команды:\n'
                           '/help — Помощь\n'
                           '/stop — Остановить и не получать больше сообщения')


@handlers.message(commands=['help'],
                  filters=[PrivateChatFilter(), FromUserFilter()])
async def user_help_command(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    assert update.message.from_ is not None
    logger.info('Help command from "%s"', update.message.from_.to_dict())
    await bot.send_message(update.message.chat.id,
                           'Пришлите сообщение или задайте вопрос. '
                           'Также вы можете использовать следующие команды:\n'
                           '/help — Помощь\n'
                           '/stop — Остановить и не получать больше сообщения')


@handlers.message(commands=['stop'],
                  filters=[PrivateChatFilter(), FromUserFilter()])
async def user_stop_command(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    assert update.message.from_ is not None
    logger.info('Stop command from "%s"', update.message.from_.to_dict())
    stopped = Stopped()
    await stopped.set(bot, update.message.from_.id)
    await remove_chat_from_list(bot, update.message.from_.id)
    group_chat = await get_chat(bot, GROUP_CHAT_KEY)
    if group_chat is None:
        notify_chat_id = await bot.storage.get(ADMIN_CHAT_ID_KEY)
        assert isinstance(notify_chat_id, int)
    else:
        notify_chat_id = group_chat.id
    await bot.send_message(
        notify_chat_id,
        f'{user_link(update.message.from_)} меня заблокировал '
        f'{stopped.dt:%Y-%m-%d %H:%M:%S %Z}.',
        parse_mode=ParseMode.HTML)
    current_chat = await get_chat(bot, CURRENT_CHAT_KEY)
    if current_chat is not None and current_chat.id == update.message.from_.id:
        await bot.storage.set(WAIT_REPLY_FROM_ID_KEY)
        await set_chat(bot, CURRENT_CHAT_KEY)


@handlers.message(commands=['start'],
                  filters=[PrivateChatFilter(), FromAdminFilter()])
async def admin_start_command(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    logger.info('Start command from admin')
    await bot.storage.set(ADMIN_CHAT_ID_KEY, update.message.chat.id)

    await bot.send_message(update.message.chat.id,
                           '/help — Помощь\n'
                           '/reply — Ответить пользователю\n'
                           '/add_to_group — Добавить в группу\n'
                           '/remove_from_group — Удалить из группы\n'
                           '/reset — Сбросить состояние')


@handlers.message(commands=['help'],
                  filters=[PrivateChatFilter(), FromAdminFilter()])
async def admin_help_command(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    logger.info('Help command from admin')
    await bot.send_message(update.message.chat.id,
                           '/help — Помощь\n'
                           '/reply — Ответить пользователю\n'
                           '/add_to_group — Добавить в группу\n'
                           '/remove_from_group — Удалить из группы\n'
                           '/reset — Сбросить состояние')


@handlers.message(commands=['reset'],
                  filters=[PrivateChatFilter(), FromAdminFilter()])
async def admin_reset_command(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    logger.info('Reset command from admin')
    await bot.storage.set(WAIT_REPLY_FROM_ID_KEY)
    await bot.storage.set(CURRENT_CHAT_KEY)
    await bot.send_message(update.message.chat.id, 'Состояние сброшено.')


@handlers.message(commands=['add_to_group'],
                  filters=[PrivateChatFilter(), FromAdminFilter()])
async def add_to_group_command(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    assert update.message.from_ is not None

    logger.info('Add to group command from "%s"',
                update.message.from_.to_dict())
    if await get_chat(bot, GROUP_CHAT_KEY) is not None:
        logger.info('Already in group. Ignore command')
        await bot.send_message(update.message.chat.id, 'Уже в группе.')
        return

    bot_username = (await bot.get_me()).username
    link = f'tg://resolve?domain={bot_username}&startgroup=startgroup'
    await bot.send_message(
        update.message.chat.id,
        f'Для добавления в группу <a href="{link}">перейдите по ссылке</a>.',
        parse_mode=ParseMode.HTML, disable_web_page_preview=True)


@handlers.message(commands=['remove_from_group'],
                  filters=[PrivateChatFilter(), FromAdminFilter()])
async def remove_from_group_command(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    assert update.message.from_ is not None
    logger.info('Remove from group command from "%s"',
                update.message.from_.to_dict())
    group_chat = await get_chat(bot, GROUP_CHAT_KEY)
    if group_chat is None:
        logger.info('Not in group. Ignore command')
        await bot.send_message(update.message.chat.id, 'Не в группе.')
        return

    try:
        await bot.leave_chat(group_chat.id)
    except TelegramError as exception:
        logger.error('Leave chat error "%s"', exception)

    await bot.send_message(update.message.chat.id,
                           f'Удален из группы <b>{group_chat.title}</b>.',
                           parse_mode=ParseMode.HTML,
                           disable_web_page_preview=True)

    await set_chat(bot, GROUP_CHAT_KEY)
    await set_chat(bot, CURRENT_CHAT_KEY)

    logger.info('Removed from group "%s"', group_chat.to_dict())


@handlers.message(commands=['start'],
                  filters=[GroupChatFilter(), FromAdminFilter()])
async def group_start_command(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    assert update.message.from_ is not None
    logger.info('Start in group command from "%s"',
                update.message.from_.to_dict())
    if await get_chat(bot, GROUP_CHAT_KEY):
        logger.info('Attempt start in group "%s"',
                    update.message.chat.to_dict())
        return

    await set_chat(bot, GROUP_CHAT_KEY, update.message.chat)
    await set_chat(bot, CURRENT_CHAT_KEY)

    await bot.set_my_commands(
        GROUP_COMMANDS, BotCommandScopeChat(update.message.chat.id))

    admin_chat_id = await bot.storage.get(ADMIN_CHAT_ID_KEY)
    assert isinstance(admin_chat_id, int)
    await bot.send_message(
        admin_chat_id, f'Запущен в <b>{update.message.chat.title}</b>.',
        parse_mode=ParseMode.HTML, disable_web_page_preview=True)

    logger.info('Started in group "%s"', update.message.chat.to_dict())


@handlers.message(commands=['help'], filters=[GroupChatFilter()])
async def group_help_command(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    assert update.message.from_ is not None
    logger.info('Help message in group from "%s"',
                update.message.from_.to_dict())
    await bot.send_message(update.message.chat.id,
                           '/help — Помощь\n'
                           '/reply — Ответить пользователю')


@handlers.message(commands=['reply'],
                  filters=[PrivateChatFilter(), FromAdminFilter()])
async def admin_reply_command(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    assert update.message.from_ is not None
    logger.info('Reply command from admin "%s"',
                update.message.from_.to_dict())
    group_chat = await get_chat(bot, GROUP_CHAT_KEY)
    if group_chat is not None:
        await bot.send_message(
            update.message.chat.id,
            f'Принимаю сообщения в группе <b>{group_chat.title}</b>.',
            parse_mode=ParseMode.HTML
        )
        logger.debug('Ignore reply command in private chat')
        return
    if await bot.storage.get(WAIT_REPLY_FROM_ID_KEY) is not None:
        await bot.send_message(update.message.chat.id, 'Уже жду сообщение.')
        logger.debug('Already wait message. Ignore command')
        return

    await reply_menu(bot, update.message.chat.id)


@handlers.message(commands=['reply'], filters=[GroupChatFilter()])
async def group_reply_command(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    assert update.message.from_ is not None
    logger.info('Reply in group command from "%s"',
                update.message.from_.to_dict())
    group_chat = await get_chat(bot, GROUP_CHAT_KEY)
    if group_chat is not None and group_chat.id != update.message.chat.id:
        await bot.leave_chat(update.message.chat.id)
        return
    if group_chat is None:
        await bot.send_message(update.message.chat.id,
                               'Не принимаю сообщения.')
        logger.debug('Ignore reply command in group')
        return
    wait_reply_from_id = await bot.storage.get(WAIT_REPLY_FROM_ID_KEY)
    if wait_reply_from_id is not None:
        assert isinstance(wait_reply_from_id, int)
        member = await bot.get_chat_member(update.message.chat.id,
                                           wait_reply_from_id)
        member_link = (user_link(member.user) if member.user.username is None
                       else f'@{member.user.username}')
        await bot.send_message(
            update.message.chat.id,
            f'Уже жду сообщение от {member_link}.',
            parse_mode=ParseMode.HTML
        )
        logger.debug('Already wait message. Ignore command')
        return

    await reply_menu(bot, update.message.chat.id)


@handlers.message(content_types=[ContentType.NEW_CHAT_MEMBERS])
async def group_new_members(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    assert update.message.new_chat_members is not None
    logger.info('New group members message "%s"', update.message.chat)
    me = await bot.get_me()
    for user in update.message.new_chat_members:
        if user.id == me.id:
            admin_chat_id = await bot.storage.get(ADMIN_CHAT_ID_KEY)
            assert isinstance(admin_chat_id, int)
            await bot.send_message(
                admin_chat_id,
                f'Добавлен в группу <b>{update.message.chat.title}</b>.',
                parse_mode=ParseMode.HTML, disable_web_page_preview=True)
            logger.info('Bot added to grouip "%s"',
                        update.message.chat.to_dict())
            group_chat = await get_chat(bot, GROUP_CHAT_KEY)
            if (
                group_chat is not None and
                group_chat.id != update.message.chat.id
            ):
                await bot.leave_chat(update.message.chat.id)
            elif (
                group_chat is not None and
                group_chat.id == update.message.chat.id
            ):
                await bot.set_my_commands(
                    GROUP_COMMANDS,
                    BotCommandScopeChat(update.message.chat.id))
            break


@handlers.message(content_types=[ContentType.LEFT_CHAT_MEMBER])
async def group_left_member(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    assert update.message.left_chat_member is not None
    logger.info('Left group member message "%s"', update.message.to_dict())
    me = await bot.get_me()
    if update.message.left_chat_member.id == me.id:
        admin_chat_id = await bot.storage.get(ADMIN_CHAT_ID_KEY)
        assert isinstance(admin_chat_id, int)
        await bot.send_message(
            admin_chat_id,
            f'Вышел из группы <b>{update.message.chat.title}</b>.',
            parse_mode=ParseMode.HTML, disable_web_page_preview=True)
        logger.info('Leave chat "%s"', update.message.chat.title)
        group_chat = await get_chat(bot, GROUP_CHAT_KEY)
        if group_chat is not None and update.message.chat.id == group_chat.id:
            await set_chat(bot, GROUP_CHAT_KEY)
            logger.info('Forget chat "%s"', update.message.chat.title)


@handlers.message(filters=[PrivateChatFilter(), FromUserFilter()])
async def user_message(bot: Bot, update: BotUpdate) -> None:
    album_forwarder = bot.get(ALBUM_FORWARDER_KEY)
    assert isinstance(album_forwarder, AlbumForwarder)
    assert update.message is not None
    assert update.message.from_ is not None
    logger.info('Message from "%s"', update.message.from_.to_dict())
    await set_chat(bot, chat_key(update.message.chat.id), update.message.chat)
    stopped = await Stopped.get(bot, update.message.chat.id)
    if stopped is not None:
        await Stopped.delete(bot, update.message.chat.id)
        await bot.send_message(update.message.chat.id, 'С возвращением!')

    group_chat = await get_chat(bot, GROUP_CHAT_KEY)
    if group_chat is None:
        forward_chat_id = await bot.storage.get(ADMIN_CHAT_ID_KEY)
        assert isinstance(forward_chat_id, int)
    else:
        forward_chat_id = group_chat.id

    if update.message.audio is not None or update.message.sticker is not None:
        logger.info('Message from user "%s" contains audio or sticker',
                    update.message.from_.to_dict())
        await send_from_message(bot, forward_chat_id, update.message.chat)

    if update.message.media_group_id is not None:
        await album_forwarder.add_message(
            update.message, forward_chat_id, add_from_info=True)
    else:
        await bot.forward_message(forward_chat_id, update.message.chat.id,
                                  update.message.message_id)

    await add_chat_to_list(bot, update.message.chat)


@handlers.message(filters=[GroupChatFilter()])
async def group_message(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    assert update.message.from_ is not None
    logger.info('Reply messgae in group from "%s"',
                update.message.from_.to_dict())
    group_chat = await get_chat(bot, GROUP_CHAT_KEY)
    if group_chat is not None and group_chat.id != update.message.chat.id:
        await bot.leave_chat(update.message.chat.id)
        return
    wait_reply_from_id = await bot.storage.get(WAIT_REPLY_FROM_ID_KEY)
    if (
        wait_reply_from_id != update.message.from_.id and
        update.message.media_group_id is None
    ):
        logger.info('Ignore message from group "%s" user "%s"',
                    update.message.chat.title, update.message.from_.to_dict())
        return

    await send_user_message(bot, update.message)

    await bot.storage.set(WAIT_REPLY_FROM_ID_KEY)
    await set_chat(bot, CURRENT_CHAT_KEY)


@handlers.message(filters=[PrivateChatFilter(), FromAdminFilter()])
async def admin_message(bot: Bot, update: BotUpdate) -> None:
    assert update.message is not None
    logger.info('Message from admin "%s"', update.message.to_dict())
    group_chat = await get_chat(bot, GROUP_CHAT_KEY)
    if group_chat is not None:
        await bot.send_message(
            update.message.chat.id,
            f'Принимаю сообщения в группе <b>{group_chat.title}</b>.',
            parse_mode=ParseMode.HTML, disable_web_page_preview=True)
        logger.info('Ignore message in private chat with admin')
        return
    wait_reply_from_id = await bot.storage.get(WAIT_REPLY_FROM_ID_KEY)
    if wait_reply_from_id is None and update.message.media_group_id is None:
        logger.info('Ignore message from admin')
        return

    await send_user_message(bot, update.message)

    await bot.storage.set(WAIT_REPLY_FROM_ID_KEY)
    await set_chat(bot, CURRENT_CHAT_KEY)


@handlers.callback_query(data_match=REPLY_RXP)
async def reply_callback(bot: Bot, update: BotUpdate) -> None:
    assert update.callback_query is not None
    assert update.callback_query.data is not None
    assert update.callback_query.message is not None
    logger.info('Reply callback query from "%s"',
                update.callback_query.from_.to_dict())
    await bot.answer_callback_query(update.callback_query.id)

    data_match = REPLY_RXP.match(update.callback_query.data)
    assert data_match is not None, 'Reply to data not match format'
    current_chat_id = int(data_match.group(CHAT_ID_GROUP))
    current_chat = await get_chat(bot, chat_key(current_chat_id))
    if current_chat is None:
        await bot.edit_message_text(
            'Ошибка. Сообщение не отправить.',
            chat_id=update.callback_query.message.chat.id,
            message_id=update.callback_query.message.message_id)
        logger.info('Skip message sending to unknown user from "%s"',
                    update.callback_query.from_.to_dict())
        return
    stopped = await Stopped.get(bot, current_chat_id)
    if stopped is not None:
        await bot.edit_message_text(
            f'{user_link(current_chat)} меня заблокировал '
            f'{stopped.dt:%Y-%m-%d %H:%M:%S %Z}.',
            chat_id=update.callback_query.message.chat.id,
            message_id=update.callback_query.message.message_id,
            parse_mode=ParseMode.HTML)
        return
    await bot.storage.set(WAIT_REPLY_FROM_ID_KEY,
                          update.callback_query.from_.id)
    await set_chat(bot, CURRENT_CHAT_KEY, current_chat)
    await bot.edit_message_text(
        f'Введите сообщение для {user_link(current_chat)}.',
        chat_id=update.callback_query.message.chat.id,
        message_id=update.callback_query.message.message_id,
        parse_mode=ParseMode.HTML, disable_web_page_preview=True)


async def run_context(
    _: Runner, storage_path: Path, config: Config
) -> AsyncIterator[None]:
    if debug():
        asyncio.get_running_loop().slow_callback_duration = 0.01

    storage = SQLiteStorage(storage_path)
    await storage.connect()
    if await storage.get(CHAT_LIST_KEY) is None:
        await storage.set(CHAT_LIST_KEY, [])
    if await storage.get(CURRENT_CHAT_KEY) is None:
        await storage.set(CURRENT_CHAT_KEY)
    if await storage.get(ADMIN_CHAT_ID_KEY) is None:
        await storage.set(ADMIN_CHAT_ID_KEY)
    if await storage.get(GROUP_CHAT_KEY) is None:
        await storage.set(GROUP_CHAT_KEY)

    handlers.freeze()
    bot = PollBot(config.tg_token, handlers, storage)
    bot[CONFIG_KEY] = config
    await bot.start()

    album_forwarder = AlbumForwarder(bot)
    bot[ALBUM_FORWARDER_KEY] = album_forwarder
    await bot[ALBUM_FORWARDER_KEY].start()

    admin_chat_id = await bot.storage.get(ADMIN_CHAT_ID_KEY)
    group_chat = await get_chat(bot, GROUP_CHAT_KEY)
    await bot.delete_my_commands()
    await bot.set_my_commands(USER_COMMANDS, BotCommandScopeAllPrivateChats())
    if admin_chat_id is not None:
        assert isinstance(admin_chat_id, int)
        await bot.set_my_commands(
            ADMIN_COMMANDS, BotCommandScopeChat(admin_chat_id))
    if group_chat is not None:
        try:
            await bot.set_my_commands(
                GROUP_COMMANDS, BotCommandScopeChat(group_chat.id))
        except TelegramError as exception:
            if exception.error_code == 403:
                logger.info('Can\'t set commands in chat "%s": %s',
                            group_chat.id, exception.description)
            else:
                raise

    yield

    await album_forwarder.stop()
    await bot.stop()
    await storage.close()


def setup_logging() -> None:
    import os

    log_format = '%(asctime)s %(name)s %(levelname)s: %(message)s'
    if debug():
        logging.basicConfig(level=logging.DEBUG, format=log_format)
        logging.getLogger('asyncio').setLevel(logging.ERROR)
        logging.getLogger('aiosqlite').setLevel(logging.INFO)
    else:
        logging.basicConfig(level=logging.INFO, format=log_format)
    logger.info('PYTHONOPTIMIZE=%s', os.environ.get('PYTHONOPTIMIZE'))
    logger.info(SOFTWARE)


def main() -> None:
    import argparse  # isort:skip
    import os  # isort:skip
    import uvloop

    parser = argparse.ArgumentParser(description='Feedback aiotgbot bot')
    parser.add_argument('config_path', type=path, help='config path')
    parser.add_argument('storage_path', type=path, help='storage path')
    args = parser.parse_args()
    if not args.config_path.exists():
        parser.error(f'config file "{args.config_path}" does not exist')
    if not (args.storage_path.is_file() or args.storage_path.parent.is_dir()):
        parser.error(f'config file "{args.storage_path}" does not exist '
                     f'and parent path is not dir')
    if TZ_KEY not in os.environ:
        parser.error('Env var TZ is not set')
    config = Config.load(args.config_path)

    setup_logging()
    uvloop.install()
    runner = Runner(run_context, debug=debug(), storage_path=args.storage_path,
                    config=config)
    runner.run()


if __name__ == '__main__':  # pragma: nocover
    main()
